# Licensed under a 3-clause BSD style license - see PYFITS.rst

"""Convenience functions for working with FITS files.

Convenience functions
=====================

The functions in this module provide shortcuts for some of the most basic
operations on FITS files, such as reading and updating the header.  They are
included directly in the 'astropy.io.fits' namespace so that they can be used
like::

    astropy.io.fits.getheader(...)

These functions are primarily for convenience when working with FITS files in
the command-line interpreter.  If performing several operations on the same
file, such as in a script, it is better to *not* use these functions, as each
one must open and re-parse the file.  In such cases it is better to use
:func:`astropy.io.fits.open` and work directly with the
:class:`astropy.io.fits.HDUList` object and underlying HDU objects.

Several of the convenience functions, such as `getheader` and `getdata` support
special arguments for selecting which HDU to use when working with a
multi-extension FITS file.  There are a few supported argument formats for
selecting the HDU.  See the documentation for `getdata` for an
explanation of all the different formats.

.. warning::
    All arguments to convenience functions other than the filename that are
    *not* for selecting the HDU should be passed in as keyword
    arguments.  This is to avoid ambiguity and conflicts with the
    HDU arguments.  For example, to set NAXIS=1 on the Primary HDU:

    Wrong::

        astropy.io.fits.setval('myimage.fits', 'NAXIS', 1)

    The above example will try to set the NAXIS value on the first extension
    HDU to blank.  That is, the argument '1' is assumed to specify an
    HDU.

    Right::

        astropy.io.fits.setval('myimage.fits', 'NAXIS', value=1)

    This will set the NAXIS keyword to 1 on the primary HDU (the default).  To
    specify the first extension HDU use::

        astropy.io.fits.setval('myimage.fits', 'NAXIS', value=1, ext=1)

    This complexity arises out of the attempt to simultaneously support
    multiple argument formats that were used in past versions of PyFITS.
    Unfortunately, it is not possible to support all formats without
    introducing some ambiguity.  A future Astropy release may standardize
    around a single format and officially deprecate the other formats.
"""

import operator
import os
import warnings

import numpy as np

from astropy.utils.exceptions import AstropyUserWarning

from .diff import FITSDiff, HDUDiff
from .file import FILE_MODES, _File
from .hdu.base import _BaseHDU, _ValidHDU
from .hdu.hdulist import HDUList, fitsopen
from .hdu.image import ImageHDU, PrimaryHDU
from .hdu.table import BinTableHDU
from .header import Header
from .util import (
    _is_dask_array,
    _is_int,
    fileobj_closed,
    fileobj_mode,
    fileobj_name,
    path_like,
)

__all__ = [
    "getheader",
    "getdata",
    "getval",
    "setval",
    "delval",
    "writeto",
    "append",
    "update",
    "info",
    "tabledump",
    "tableload",
    "table_to_hdu",
    "printdiff",
]


def getheader(filename, *args, **kwargs):
    """
    Get the header from an HDU of a FITS file.

    Parameters
    ----------
    filename : path-like or file-like
        File to get header from.  If an opened file object, its mode
        must be one of the following rb, rb+, or ab+).

    ext, extname, extver
        The rest of the arguments are for HDU specification.  See the
        `getdata` documentation for explanations/examples.
    **kwargs
        Any additional keyword arguments to be passed to
        `astropy.io.fits.open`.

    Returns
    -------
    header : `Header` object
    """
    mode, closed = _get_file_mode(filename)
    hdulist, extidx = _getext(filename, mode, *args, **kwargs)
    try:
        hdu = hdulist[extidx]
        header = hdu.header
    finally:
        hdulist.close(closed=closed)

    return header


def getdata(filename, *args, header=None, lower=None, upper=None, view=None, **kwargs):
    """
    Get the data from an HDU of a FITS file (and optionally the
    header).

    Parameters
    ----------
    filename : path-like or file-like
        File to get data from.  If opened, mode must be one of the
        following rb, rb+, or ab+.

    ext
        The rest of the arguments are for HDU specification.
        They are flexible and are best illustrated by examples.

        No extra arguments implies the primary HDU::

            getdata('in.fits')

        .. note::
            Exclusive to ``getdata``: if ``ext`` is not specified
            and primary header contains no data, ``getdata`` attempts
            to retrieve data from first extension HDU.

        By HDU number::

            getdata('in.fits', 0)      # the primary HDU
            getdata('in.fits', 2)      # the second extension HDU
            getdata('in.fits', ext=2)  # the second extension HDU

        By name, i.e., ``EXTNAME`` value (if unique)::

            getdata('in.fits', 'sci')
            getdata('in.fits', extname='sci')  # equivalent

        Note ``EXTNAME`` values are not case sensitive

        By combination of ``EXTNAME`` and EXTVER`` as separate
        arguments or as a tuple::

            getdata('in.fits', 'sci', 2)  # EXTNAME='SCI' & EXTVER=2
            getdata('in.fits', extname='sci', extver=2)  # equivalent
            getdata('in.fits', ('sci', 2))  # equivalent

        Ambiguous or conflicting specifications will raise an exception::

            getdata('in.fits', ext=('sci',1), extname='err', extver=2)

    header : bool, optional
        If `True`, return the data and the header of the specified HDU as a
        tuple.

    lower, upper : bool, optional
        If ``lower`` or ``upper`` are `True`, the field names in the
        returned data object will be converted to lower or upper case,
        respectively.

    view : ndarray, optional
        When given, the data will be returned wrapped in the given ndarray
        subclass by calling::

           data.view(view)

    **kwargs
        Any additional keyword arguments to be passed to
        `astropy.io.fits.open`.

    Returns
    -------
    array : ndarray or `~numpy.recarray` or `~astropy.io.fits.Group`
        Type depends on the type of the extension being referenced.

        If the optional keyword ``header`` is set to `True`, this
        function will return a (``data``, ``header``) tuple.

    Raises
    ------
    IndexError
        If no data is found in searched HDUs.
    """
    mode, closed = _get_file_mode(filename)

    ext = kwargs.get("ext")
    extname = kwargs.get("extname")
    extver = kwargs.get("extver")
    ext_given = not (
        len(args) == 0 and ext is None and extname is None and extver is None
    )

    hdulist, extidx = _getext(filename, mode, *args, **kwargs)
    try:
        hdu = hdulist[extidx]
        data = hdu.data
        if data is None:
            if ext_given:
                raise IndexError(f"No data in HDU #{extidx}.")

            # fallback to the first extension HDU
            if len(hdulist) == 1:
                raise IndexError("No data in Primary HDU and no extension HDU found.")
            hdu = hdulist[1]
            data = hdu.data
            if data is None:
                raise IndexError("No data in either Primary or first extension HDUs.")

        if header:
            hdr = hdu.header
    finally:
        hdulist.close(closed=closed)

    # Change case of names if requested
    trans = None
    if lower:
        trans = operator.methodcaller("lower")
    elif upper:
        trans = operator.methodcaller("upper")
    if trans:
        if data.dtype.names is None:
            # this data does not have fields
            return
        if data.dtype.descr[0][0] == "":
            # this data does not have fields
            return
        data.dtype.names = [trans(n) for n in data.dtype.names]

    # allow different views into the underlying ndarray.  Keep the original
    # view just in case there is a problem
    if isinstance(view, type) and issubclass(view, np.ndarray):
        data = data.view(view)

    if header:
        return data, hdr
    else:
        return data


def getval(filename, keyword, *args, **kwargs):
    """
    Get a keyword's value from a header in a FITS file.

    Parameters
    ----------
    filename : path-like or file-like
        Name of the FITS file, or file object (if opened, mode must be
        one of the following rb, rb+, or ab+).

    keyword : str
        Keyword name

    ext, extname, extver
        The rest of the arguments are for HDU specification.
        See `getdata` for explanations/examples.
    **kwargs
        Any additional keyword arguments to be passed to
        `astropy.io.fits.open`.
        *Note:* This function automatically specifies ``do_not_scale_image_data
        = True`` when opening the file so that values can be retrieved from the
        unmodified header.

    Returns
    -------
    keyword value : str, int, or float
    """
    if "do_not_scale_image_data" not in kwargs:
        kwargs["do_not_scale_image_data"] = True

    hdr = getheader(filename, *args, **kwargs)
    return hdr[keyword]


def setval(
    filename,
    keyword,
    *args,
    value=None,
    comment=None,
    before=None,
    after=None,
    savecomment=False,
    **kwargs,
):
    """
    Set a keyword's value from a header in a FITS file.

    If the keyword already exists, it's value/comment will be updated.
    If it does not exist, a new card will be created and it will be
    placed before or after the specified location.  If no ``before`` or
    ``after`` is specified, it will be appended at the end.

    When updating more than one keyword in a file, this convenience
    function is a much less efficient approach compared with opening
    the file for update, modifying the header, and closing the file.

    Parameters
    ----------
    filename : path-like or file-like
        Name of the FITS file, or file object If opened, mode must be update
        (rb+).  An opened file object or `~gzip.GzipFile` object will be closed
        upon return.

    keyword : str
        Keyword name

    value : str, int, float, optional
        Keyword value (default: `None`, meaning don't modify)

    comment : str, optional
        Keyword comment, (default: `None`, meaning don't modify)

    before : str, int, optional
        Name of the keyword, or index of the card before which the new card
        will be placed.  The argument ``before`` takes precedence over
        ``after`` if both are specified (default: `None`).

    after : str, int, optional
        Name of the keyword, or index of the card after which the new card will
        be placed. (default: `None`).

    savecomment : bool, optional
        When `True`, preserve the current comment for an existing keyword.  The
        argument ``savecomment`` takes precedence over ``comment`` if both
        specified.  If ``comment`` is not specified then the current comment
        will automatically be preserved  (default: `False`).

    ext, extname, extver
        The rest of the arguments are for HDU specification.
        See `getdata` for explanations/examples.
    **kwargs
        Any additional keyword arguments to be passed to
        `astropy.io.fits.open`.
        *Note:* This function automatically specifies ``do_not_scale_image_data
        = True`` when opening the file so that values can be retrieved from the
        unmodified header.
    """
    if "do_not_scale_image_data" not in kwargs:
        kwargs["do_not_scale_image_data"] = True

    closed = fileobj_closed(filename)
    hdulist, extidx = _getext(filename, "update", *args, **kwargs)
    try:
        if keyword in hdulist[extidx].header and savecomment:
            comment = None
        hdulist[extidx].header.set(keyword, value, comment, before, after)
    finally:
        hdulist.close(closed=closed)


def delval(filename, keyword, *args, **kwargs):
    """
    Delete all instances of keyword from a header in a FITS file.

    Parameters
    ----------
    filename : path-like or file-like
        Name of the FITS file, or file object If opened, mode must be update
        (rb+).  An opened file object or `~gzip.GzipFile` object will be closed
        upon return.

    keyword : str, int
        Keyword name or index

    ext, extname, extver
        The rest of the arguments are for HDU specification.
        See `getdata` for explanations/examples.
    **kwargs
        Any additional keyword arguments to be passed to
        `astropy.io.fits.open`.
        *Note:* This function automatically specifies ``do_not_scale_image_data
        = True`` when opening the file so that values can be retrieved from the
        unmodified header.
    """
    if "do_not_scale_image_data" not in kwargs:
        kwargs["do_not_scale_image_data"] = True

    closed = fileobj_closed(filename)
    hdulist, extidx = _getext(filename, "update", *args, **kwargs)
    try:
        del hdulist[extidx].header[keyword]
    finally:
        hdulist.close(closed=closed)


def writeto(
    filename,
    data,
    header=None,
    output_verify="exception",
    overwrite=False,
    checksum=False,
):
    """
    Create a new FITS file using the supplied data/header.

    Parameters
    ----------
    filename : path-like or file-like
        File to write to.  If opened, must be opened in a writable binary
        mode such as 'wb' or 'ab+'.

    data : array or `~numpy.recarray` or `~astropy.io.fits.Group`
        data to write to the new file

    header : `Header` object, optional
        the header associated with ``data``. If `None`, a header
        of the appropriate type is created for the supplied data. This
        argument is optional.

    output_verify : str
        Output verification option.  Must be one of ``"fix"``, ``"silentfix"``,
        ``"ignore"``, ``"warn"``, or ``"exception"``.  May also be any
        combination of ``"fix"`` or ``"silentfix"`` with ``"+ignore"``,
        ``+warn``, or ``+exception" (e.g. ``"fix+warn"``).  See
        :ref:`astropy:verify` for more info.

    overwrite : bool, optional
        If ``True``, overwrite the output file if it exists. Raises an
        ``OSError`` if ``False`` and the output file exists. Default is
        ``False``.

    checksum : bool, optional
        If `True`, adds both ``DATASUM`` and ``CHECKSUM`` cards to the
        headers of all HDU's written to the file.

    Notes
    -----
    gzip, zip and bzip2 compression algorithms are natively supported.
    Compression mode is determined from the filename extension
    ('.gz', '.zip' or '.bz2' respectively).  It is also possible to pass a
    compressed file object, e.g. `gzip.GzipFile`.
    """
    hdu = _makehdu(data, header)
    if hdu.is_image and not isinstance(hdu, PrimaryHDU):
        hdu = PrimaryHDU(data, header=header)
    hdu.writeto(
        filename, overwrite=overwrite, output_verify=output_verify, checksum=checksum
    )


def table_to_hdu(table, character_as_bytes=False):
    """
    Convert an `~astropy.table.Table` object to a FITS
    `~astropy.io.fits.BinTableHDU`.

    Parameters
    ----------
    table : astropy.table.Table
        The table to convert.
    character_as_bytes : bool
        Whether to return bytes for string columns when accessed from the HDU.
        By default this is `False` and (unicode) strings are returned, but for
        large tables this may use up a lot of memory.

    Returns
    -------
    table_hdu : `~astropy.io.fits.BinTableHDU`
        The FITS binary table HDU.
    """
    # Avoid circular imports
    from .column import python_to_tdisp
    from .connect import REMOVE_KEYWORDS, is_column_keyword

    # Header to store Time related metadata
    hdr = None

    # Not all tables with mixin columns are supported
    if table.has_mixin_columns:
        # Import is done here, in order to avoid it at build time as erfa is not
        # yet available then.
        from astropy.table.column import BaseColumn
        from astropy.time import Time
        from astropy.units import Quantity

        from .fitstime import time_to_fits

        # Only those columns which are instances of BaseColumn, Quantity or Time can
        # be written
        unsupported_cols = table.columns.not_isinstance((BaseColumn, Quantity, Time))
        if unsupported_cols:
            unsupported_names = [col.info.name for col in unsupported_cols]
            raise ValueError(
                f"cannot write table with mixin column(s) {unsupported_names}"
            )

        time_cols = table.columns.isinstance(Time)
        if time_cols:
            table, hdr = time_to_fits(table)

    # Create a new HDU object
    tarray = table.as_array()
    if isinstance(tarray, np.ma.MaskedArray):
        # Fill masked values carefully:
        # float column's default mask value needs to be Nan and
        # string column's default mask should be an empty string.
        # Note: getting the fill value for the structured array is
        # more reliable than for individual columns for string entries.
        # (no 'N/A' for a single-element string, where it should be 'N').
        default_fill_value = np.ma.default_fill_value(tarray.dtype)
        for colname, (coldtype, _) in tarray.dtype.fields.items():
            if np.all(tarray.fill_value[colname] == default_fill_value[colname]):
                # Since multi-element columns with dtypes such as '2f8' have
                # a subdtype, we should look up the type of column on that.
                coltype = (
                    coldtype.subdtype[0].type if coldtype.subdtype else coldtype.type
                )
                if issubclass(coltype, np.complexfloating):
                    tarray.fill_value[colname] = complex(np.nan, np.nan)
                elif issubclass(coltype, np.inexact):
                    tarray.fill_value[colname] = np.nan
                elif issubclass(coltype, np.character):
                    tarray.fill_value[colname] = ""

        # TODO: it might be better to construct the FITS table directly from
        # the Table columns, rather than go via a structured array.
        table_hdu = BinTableHDU.from_columns(
            tarray.filled(), header=hdr, character_as_bytes=character_as_bytes
        )
        for col in table_hdu.columns:
            # Binary FITS tables support TNULL *only* for integer data columns
            # TODO: Determine a schema for handling non-integer masked columns
            # with non-default fill values in FITS (if at all possible).
            # Be careful that we do not set null for columns that were not masked!
            int_formats = ("B", "I", "J", "K")
            if (
                col.format in int_formats or col.format.p_format in int_formats
            ) and hasattr(table[col.name], "mask"):
                fill_value = tarray[col.name].fill_value
                col.null = fill_value.astype(int)
    else:
        table_hdu = BinTableHDU.from_columns(
            tarray, header=hdr, character_as_bytes=character_as_bytes
        )

    # Set units and format display for output HDU
    for col in table_hdu.columns:
        if table[col.name].info.format is not None:
            # check for boolean types, special format case
            logical = table[col.name].info.dtype == bool

            tdisp_format = python_to_tdisp(
                table[col.name].info.format, logical_dtype=logical
            )
            if tdisp_format is not None:
                col.disp = tdisp_format

        unit = table[col.name].unit
        if unit is not None:
            # Local imports to avoid importing units when it is not required,
            # e.g. for command-line scripts
            from astropy.units import Unit
            from astropy.units.format.fits import UnitScaleError

            try:
                col.unit = unit.to_string(format="fits")
            except UnitScaleError:
                scale = unit.scale
                raise UnitScaleError(
                    f"The column '{col.name}' could not be stored in FITS "
                    f"format because it has a scale '({str(scale)})' that "
                    "is not recognized by the FITS standard. Either scale "
                    "the data or change the units."
                )
            except ValueError:
                # Warn that the unit is lost, but let the details depend on
                # whether the column was serialized (because it was a
                # quantity), since then the unit can be recovered by astropy.
                warning = (
                    f"The unit '{unit.to_string()}' could not be saved in "
                    "native FITS format "
                )
                if any(
                    "SerializedColumn" in item and "name: " + col.name in item
                    for item in table.meta.get("comments", [])
                ):
                    warning += (
                        "and hence will be lost to non-astropy fits readers. "
                        "Within astropy, the unit can roundtrip using QTable, "
                        "though one has to enable the unit before reading."
                    )
                else:
                    warning += (
                        "and cannot be recovered in reading. It can roundtrip "
                        "within astropy by using QTable both to write and read "
                        "back, though one has to enable the unit before reading."
                    )
                warnings.warn(warning, AstropyUserWarning)

            else:
                # Try creating a Unit to issue a warning if the unit is not
                # FITS compliant
                Unit(col.unit, format="fits", parse_strict="warn")

    # Column-specific override keywords for coordinate columns
    coord_meta = table.meta.pop("__coordinate_columns__", {})
    for col_name, col_info in coord_meta.items():
        col = table_hdu.columns[col_name]
        # Set the column coordinate attributes from data saved earlier.
        # Note: have to set these, even if we have no data.
        for attr in "coord_type", "coord_unit":
            setattr(col, attr, col_info.get(attr, None))
        trpos = col_info.get("time_ref_pos", None)
        if trpos is not None:
            col.time_ref_pos = trpos

    for key, value in table.meta.items():
        if is_column_keyword(key.upper()) or key.upper() in REMOVE_KEYWORDS:
            warnings.warn(
                f"Meta-data keyword {key} will be ignored since it conflicts "
                "with a FITS reserved keyword",
                AstropyUserWarning,
            )
            continue

        # Convert to FITS format
        if key == "comments":
            key = "comment"

        if isinstance(value, list):
            for item in value:
                try:
                    table_hdu.header.append((key, item))
                except ValueError:
                    warnings.warn(
                        f"Attribute `{key}` of type {type(value)} cannot be "
                        "added to FITS Header - skipping",
                        AstropyUserWarning,
                    )
        else:
            try:
                table_hdu.header[key] = value
            except ValueError:
                warnings.warn(
                    f"Attribute `{key}` of type {type(value)} cannot be "
                    "added to FITS Header - skipping",
                    AstropyUserWarning,
                )
    return table_hdu


def append(filename, data, header=None, checksum=False, verify=True, **kwargs):
    """
    Append the header/data to FITS file if filename exists, create if not.

    If only ``data`` is supplied, a minimal header is created.

    Parameters
    ----------
    filename : path-like or file-like
        File to write to.  If opened, must be opened for update (rb+) unless it
        is a new file, then it must be opened for append (ab+).  A file or
        `~gzip.GzipFile` object opened for update will be closed after return.

    data : array, :class:`~astropy.table.Table`, or `~astropy.io.fits.Group`
        The new data used for appending.

    header : `Header` object, optional
        The header associated with ``data``.  If `None`, an appropriate header
        will be created for the data object supplied.

    checksum : bool, optional
        When `True` adds both ``DATASUM`` and ``CHECKSUM`` cards to the header
        of the HDU when written to the file.

    verify : bool, optional
        When `True`, the existing FITS file will be read in to verify it for
        correctness before appending.  When `False`, content is simply appended
        to the end of the file.  Setting ``verify`` to `False` can be much
        faster.

    **kwargs
        Additional arguments are passed to:

        - `~astropy.io.fits.writeto` if the file does not exist or is empty.
          In this case ``output_verify`` is the only possible argument.
        - `~astropy.io.fits.open` if ``verify`` is True or if ``filename``
          is a file object.
        - Otherwise no additional arguments can be used.

    """
    if isinstance(filename, path_like):
        filename = os.path.expanduser(filename)
    name, closed, noexist_or_empty = _stat_filename_or_fileobj(filename)

    if noexist_or_empty:
        #
        # The input file or file like object either doesn't exits or is
        # empty.  Use the writeto convenience function to write the
        # output to the empty object.
        #
        writeto(filename, data, header, checksum=checksum, **kwargs)
    else:
        hdu = _makehdu(data, header)

        if isinstance(hdu, PrimaryHDU):
            hdu = ImageHDU(data, header)

        if verify or not closed:
            f = fitsopen(filename, mode="append", **kwargs)
            try:
                f.append(hdu)

                # Set a flag in the HDU so that only this HDU gets a checksum
                # when writing the file.
                hdu._output_checksum = checksum
            finally:
                f.close(closed=closed)
        else:
            f = _File(filename, mode="append")
            try:
                hdu._output_checksum = checksum
                hdu._writeto(f)
            finally:
                f.close()


def update(filename, data, *args, **kwargs):
    """
    Update the specified HDU with the input data/header.

    Parameters
    ----------
    filename : path-like or file-like
        File to update.  If opened, mode must be update (rb+).  An opened file
        object or `~gzip.GzipFile` object will be closed upon return.

    data : array, `~astropy.table.Table`, or `~astropy.io.fits.Group`
        The new data used for updating.

    header : `Header` object, optional
        The header associated with ``data``.  If `None`, an appropriate header
        will be created for the data object supplied.

    ext, extname, extver
        The rest of the arguments are flexible: the 3rd argument can be the
        header associated with the data.  If the 3rd argument is not a
        `Header`, it (and other positional arguments) are assumed to be the
        HDU specification(s).  Header and HDU specs can also be
        keyword arguments.  For example::

            update(file, dat, hdr, 'sci')  # update the 'sci' extension
            update(file, dat, 3)  # update the 3rd extension HDU
            update(file, dat, hdr, 3)  # update the 3rd extension HDU
            update(file, dat, 'sci', 2)  # update the 2nd extension HDU named 'sci'
            update(file, dat, 3, header=hdr)  # update the 3rd extension HDU
            update(file, dat, header=hdr, ext=5)  # update the 5th extension HDU

    **kwargs
        Any additional keyword arguments to be passed to
        `astropy.io.fits.open`.
    """
    # The arguments to this function are a bit trickier to deal with than others
    # in this module, since the documentation has promised that the header
    # argument can be an optional positional argument.
    if args and isinstance(args[0], Header):
        header = args[0]
        args = args[1:]
    else:
        header = None
    # The header can also be a keyword argument--if both are provided the
    # keyword takes precedence
    header = kwargs.pop("header", header)

    new_hdu = _makehdu(data, header)

    closed = fileobj_closed(filename)

    hdulist, _ext = _getext(filename, "update", *args, **kwargs)
    try:
        hdulist[_ext] = new_hdu
    finally:
        hdulist.close(closed=closed)


def info(filename, output=None, **kwargs):
    """
    Print the summary information on a FITS file.

    This includes the name, type, length of header, data shape and type
    for each HDU.

    Parameters
    ----------
    filename : path-like or file-like
        FITS file to obtain info from.  If opened, mode must be one of
        the following: rb, rb+, or ab+ (i.e. the file must be readable).

    output : file, bool, optional
        A file-like object to write the output to.  If ``False``, does not
        output to a file and instead returns a list of tuples representing the
        HDU info.  Writes to ``sys.stdout`` by default.
    **kwargs
        Any additional keyword arguments to be passed to
        `astropy.io.fits.open`.
        *Note:* This function sets ``ignore_missing_end=True`` by default.
    """
    mode, closed = _get_file_mode(filename, default="readonly")
    # Set the default value for the ignore_missing_end parameter
    if "ignore_missing_end" not in kwargs:
        kwargs["ignore_missing_end"] = True

    f = fitsopen(filename, mode=mode, **kwargs)
    try:
        ret = f.info(output=output)
    finally:
        if closed:
            f.close()

    return ret


def printdiff(inputa, inputb, *args, **kwargs):
    """
    Compare two parts of a FITS file, including entire FITS files,
    FITS `HDUList` objects and FITS ``HDU`` objects.

    Parameters
    ----------
    inputa : str, `HDUList` object, or ``HDU`` object
        The filename of a FITS file, `HDUList`, or ``HDU``
        object to compare to ``inputb``.

    inputb : str, `HDUList` object, or ``HDU`` object
        The filename of a FITS file, `HDUList`, or ``HDU``
        object to compare to ``inputa``.

    ext, extname, extver
        Additional positional arguments are for HDU specification if your
        inputs are string filenames (will not work if
        ``inputa`` and ``inputb`` are ``HDU`` objects or `HDUList` objects).
        They are flexible and are best illustrated by examples.  In addition
        to using these arguments positionally you can directly call the
        keyword parameters ``ext``, ``extname``.

        By HDU number::

            printdiff('inA.fits', 'inB.fits', 0)      # the primary HDU
            printdiff('inA.fits', 'inB.fits', 2)      # the second extension HDU
            printdiff('inA.fits', 'inB.fits', ext=2)  # the second extension HDU

        By name, i.e., ``EXTNAME`` value (if unique). ``EXTNAME`` values are
        not case sensitive:

            printdiff('inA.fits', 'inB.fits', 'sci')
            printdiff('inA.fits', 'inB.fits', extname='sci')  # equivalent

        By combination of ``EXTNAME`` and ``EXTVER`` as separate
        arguments or as a tuple::

            printdiff('inA.fits', 'inB.fits', 'sci', 2)    # EXTNAME='SCI'
                                                           # & EXTVER=2
            printdiff('inA.fits', 'inB.fits', extname='sci', extver=2)
                                                           # equivalent
            printdiff('inA.fits', 'inB.fits', ('sci', 2))  # equivalent

        Ambiguous or conflicting specifications will raise an exception::

            printdiff('inA.fits', 'inB.fits',
                      ext=('sci', 1), extname='err', extver=2)

    **kwargs
        Any additional keyword arguments to be passed to
        `~astropy.io.fits.FITSDiff`.

    Notes
    -----
    The primary use for the `printdiff` function is to allow quick print out
    of a FITS difference report and will write to ``sys.stdout``.
    To save the diff report to a file please use `~astropy.io.fits.FITSDiff`
    directly.
    """
    # Pop extension keywords
    extension = {
        key: kwargs.pop(key) for key in ["ext", "extname", "extver"] if key in kwargs
    }
    has_extensions = args or extension

    if isinstance(inputa, str) and has_extensions:
        # Use handy _getext to interpret any ext keywords, but
        # will need to close a if  fails
        modea, closeda = _get_file_mode(inputa)
        modeb, closedb = _get_file_mode(inputb)

        hdulista, extidxa = _getext(inputa, modea, *args, **extension)
        # Have to close a if b doesn't make it
        try:
            hdulistb, extidxb = _getext(inputb, modeb, *args, **extension)
        except Exception:
            hdulista.close(closed=closeda)
            raise

        try:
            hdua = hdulista[extidxa]
            hdub = hdulistb[extidxb]
            # See below print for note
            print(HDUDiff(hdua, hdub, **kwargs).report())

        finally:
            hdulista.close(closed=closeda)
            hdulistb.close(closed=closedb)

    # If input is not a string, can feed HDU objects or HDUList directly,
    # but can't currently handle extensions
    elif isinstance(inputa, _ValidHDU) and has_extensions:
        raise ValueError("Cannot use extension keywords when providing an HDU object.")

    elif isinstance(inputa, _ValidHDU) and not has_extensions:
        print(HDUDiff(inputa, inputb, **kwargs).report())

    elif isinstance(inputa, HDUList) and has_extensions:
        raise NotImplementedError(
            "Extension specification with HDUList objects not implemented."
        )

    # This function is EXCLUSIVELY for printing the diff report to screen
    # in a one-liner call, hence the use of print instead of logging
    else:
        print(FITSDiff(inputa, inputb, **kwargs).report())


def tabledump(filename, datafile=None, cdfile=None, hfile=None, ext=1, overwrite=False):
    """
    Dump a table HDU to a file in ASCII format.  The table may be
    dumped in three separate files, one containing column definitions,
    one containing header parameters, and one for table data.

    Parameters
    ----------
    filename : path-like or file-like
        Input fits file.

    datafile : path-like or file-like, optional
        Output data file.  The default is the root name of the input
        fits file appended with an underscore, followed by the
        extension number (ext), followed by the extension ``.txt``.

    cdfile : path-like or file-like, optional
        Output column definitions file.  The default is `None`,
        no column definitions output is produced.

    hfile : path-like or file-like, optional
        Output header parameters file.  The default is `None`,
        no header parameters output is produced.

    ext : int
        The number of the extension containing the table HDU to be
        dumped.

    overwrite : bool, optional
        If ``True``, overwrite the output file if it exists. Raises an
        ``OSError`` if ``False`` and the output file exists. Default is
        ``False``.

    Notes
    -----
    The primary use for the `tabledump` function is to allow editing in a
    standard text editor of the table data and parameters.  The
    `tableload` function can be used to reassemble the table from the
    three ASCII files.
    """
    # allow file object to already be opened in any of the valid modes
    # and leave the file in the same state (opened or closed) as when
    # the function was called

    mode, closed = _get_file_mode(filename, default="readonly")
    f = fitsopen(filename, mode=mode)

    # Create the default data file name if one was not provided
    try:
        if not datafile:
            root, tail = os.path.splitext(f._file.name)
            datafile = root + "_" + repr(ext) + ".txt"

        # Dump the data from the HDU to the files
        f[ext].dump(datafile, cdfile, hfile, overwrite)
    finally:
        if closed:
            f.close()


if isinstance(tabledump.__doc__, str):
    tabledump.__doc__ += BinTableHDU._tdump_file_format.replace("\n", "\n    ")


def tableload(datafile, cdfile, hfile=None):
    """
    Create a table from the input ASCII files.  The input is from up
    to three separate files, one containing column definitions, one
    containing header parameters, and one containing column data.  The
    header parameters file is not required.  When the header
    parameters file is absent a minimal header is constructed.

    Parameters
    ----------
    datafile : path-like or file-like
        Input data file containing the table data in ASCII format.

    cdfile : path-like or file-like
        Input column definition file containing the names, formats,
        display formats, physical units, multidimensional array
        dimensions, undefined values, scale factors, and offsets
        associated with the columns in the table.

    hfile : path-like or file-like, optional
        Input parameter definition file containing the header
        parameter definitions to be associated with the table.
        If `None`, a minimal header is constructed.

    Notes
    -----
    The primary use for the `tableload` function is to allow the input of
    ASCII data that was edited in a standard text editor of the table
    data and parameters.  The tabledump function can be used to create the
    initial ASCII files.
    """
    return BinTableHDU.load(datafile, cdfile, hfile, replace=True)


if isinstance(tableload.__doc__, str):
    tableload.__doc__ += BinTableHDU._tdump_file_format.replace("\n", "\n    ")


def _getext(filename, mode, *args, ext=None, extname=None, extver=None, **kwargs):
    """
    Open the input file, return the `HDUList` and the extension.

    This supports several different styles of extension selection.  See the
    :func:`getdata()` documentation for the different possibilities.
    """
    err_msg = "Redundant/conflicting extension arguments(s): {}".format(
        {"args": args, "ext": ext, "extname": extname, "extver": extver}
    )

    # This code would be much simpler if just one way of specifying an
    # extension were picked.  But now we need to support all possible ways for
    # the time being.
    if len(args) == 1:
        # Must be either an extension number, an extension name, or an
        # (extname, extver) tuple
        if _is_int(args[0]) or (isinstance(ext, tuple) and len(ext) == 2):
            if ext is not None or extname is not None or extver is not None:
                raise TypeError(err_msg)
            ext = args[0]
        elif isinstance(args[0], str):
            # The first arg is an extension name; it could still be valid
            # to provide an extver kwarg
            if ext is not None or extname is not None:
                raise TypeError(err_msg)
            extname = args[0]
        else:
            # Take whatever we have as the ext argument; we'll validate it
            # below
            ext = args[0]
    elif len(args) == 2:
        # Must be an extname and extver
        if ext is not None or extname is not None or extver is not None:
            raise TypeError(err_msg)
        extname = args[0]
        extver = args[1]
    elif len(args) > 2:
        raise TypeError("Too many positional arguments.")

    if ext is not None and not (
        _is_int(ext)
        or (
            isinstance(ext, tuple)
            and len(ext) == 2
            and isinstance(ext[0], str)
            and _is_int(ext[1])
        )
    ):
        raise ValueError(
            "The ext keyword must be either an extension number "
            "(zero-indexed) or a (extname, extver) tuple."
        )
    if extname is not None and not isinstance(extname, str):
        raise ValueError("The extname argument must be a string.")
    if extver is not None and not _is_int(extver):
        raise ValueError("The extver argument must be an integer.")

    if ext is None and extname is None and extver is None:
        ext = 0
    elif ext is not None and (extname is not None or extver is not None):
        raise TypeError(err_msg)
    elif extname:
        if extver:
            ext = (extname, extver)
        else:
            ext = (extname, 1)
    elif extver and extname is None:
        raise TypeError("extver alone cannot specify an extension.")

    hdulist = fitsopen(filename, mode=mode, **kwargs)

    return hdulist, ext


def _makehdu(data, header):
    if header is None:
        header = Header()
    hdu = _BaseHDU._from_data(data, header)
    if hdu.__class__ in (_BaseHDU, _ValidHDU):
        # The HDU type was unrecognized, possibly due to a
        # nonexistent/incomplete header
        if (
            isinstance(data, np.ndarray) and data.dtype.fields is not None
        ) or isinstance(data, np.recarray):
            hdu = BinTableHDU(data, header=header)
        elif isinstance(data, np.ndarray) or _is_dask_array(data):
            hdu = ImageHDU(data, header=header)
        else:
            raise KeyError("Data must be a numpy array.")
    return hdu


def _stat_filename_or_fileobj(filename):
    if isinstance(filename, os.PathLike):
        filename = os.fspath(filename)
    closed = fileobj_closed(filename)
    name = fileobj_name(filename) or ""

    try:
        loc = filename.tell()
    except AttributeError:
        loc = 0

    noexist_or_empty = (
        name and (not os.path.exists(name) or (os.path.getsize(name) == 0))
    ) or (not name and loc == 0)

    return name, closed, noexist_or_empty


def _get_file_mode(filename, default="readonly"):
    """
    Allow file object to already be opened in any of the valid modes and
    and leave the file in the same state (opened or closed) as when
    the function was called.
    """
    mode = default
    closed = fileobj_closed(filename)

    fmode = fileobj_mode(filename)
    if fmode is not None:
        mode = FILE_MODES.get(fmode)
        if mode is None:
            raise OSError(
                f"File mode of the input file object ({fmode!r}) cannot be used to "
                "read/write FITS files."
            )

    return mode, closed



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pytest

from astropy import __version__ as version
from astropy.io.fits.scripts import fitsheader

from .conftest import FitsTestCase


class TestFITSheader_script(FitsTestCase):
    def test_help(self):
        with pytest.raises(SystemExit) as e:
            fitsheader.main(["-h"])
        assert e.value.code == 0

    def test_version(self, capsys):
        with pytest.raises(SystemExit) as e:
            fitsheader.main(["--version"])
            out = capsys.readouterr()[0]
            assert out == f"fitsheader {version}"
        assert e.value.code == 0

    def test_file_exists(self, capsys):
        fitsheader.main([self.data("arange.fits")])
        out, err = capsys.readouterr()
        assert out.splitlines()[1].startswith(
            "SIMPLE  =                    T / conforms to FITS standard"
        )
        assert err == ""

    def test_by_keyword(self, capsys):
        fitsheader.main(["-k", "NAXIS", self.data("arange.fits")])
        out, err = capsys.readouterr()
        assert out.splitlines()[1].startswith(
            "NAXIS   =                    3 / number of array dimensions"
        )

        fitsheader.main(["-k", "NAXIS*", self.data("arange.fits")])
        out, err = capsys.readouterr()
        out = out.splitlines()
        assert len(out) == 5
        assert out[1].startswith("NAXIS")
        assert out[2].startswith("NAXIS1")
        assert out[3].startswith("NAXIS2")
        assert out[4].startswith("NAXIS3")

        fitsheader.main(["-k", "RANDOMKEY", self.data("arange.fits")])
        out, err = capsys.readouterr()
        assert err.startswith("WARNING") and "RANDOMKEY" in err
        assert not err.startswith("ERROR")

    def test_by_extension(self, capsys):
        fitsheader.main(["-e", "1", self.data("test0.fits")])
        out, err = capsys.readouterr()
        assert len(out.splitlines()) == 62

        fitsheader.main(["-e", "3", "-k", "BACKGRND", self.data("test0.fits")])
        out, err = capsys.readouterr()
        assert out.splitlines()[1].startswith("BACKGRND=                 312.")

        fitsheader.main(["-e", "0", "-k", "BACKGRND", self.data("test0.fits")])
        out, err = capsys.readouterr()
        assert err.startswith("WARNING")

        fitsheader.main(["-e", "3", "-k", "FOO", self.data("test0.fits")])
        out, err = capsys.readouterr()
        assert err.startswith("WARNING")

    def test_table(self, capsys):
        fitsheader.main(["-t", "-k", "BACKGRND", self.data("test0.fits")])
        out, err = capsys.readouterr()
        out = out.splitlines()
        assert len(out) == 5
        assert out[1].endswith("|   1 | BACKGRND | 316.0 |")
        assert out[2].endswith("|   2 | BACKGRND | 351.0 |")
        assert out[3].endswith("|   3 | BACKGRND | 312.0 |")
        assert out[4].endswith("|   4 | BACKGRND | 323.0 |")

        fitsheader.main(
            [
                "-t",
                "-e",
                "0",
                "-k",
                "NAXIS",
                self.data("arange.fits"),
                self.data("ascii.fits"),
                self.data("blank.fits"),
            ]
        )
        out, err = capsys.readouterr()
        out = out.splitlines()
        assert len(out) == 4
        assert out[1].endswith("|   0 |   NAXIS |     3 |")
        assert out[2].endswith("|   0 |   NAXIS |     0 |")
        assert out[3].endswith("|   0 |   NAXIS |     2 |")

    def test_fitsort(self, capsys):
        fitsheader.main(
            [
                "-e",
                "0",
                "-f",
                "-k",
                "EXPSTART",
                "-k",
                "EXPTIME",
                self.data("test0.fits"),
                self.data("test1.fits"),
            ]
        )
        out, err = capsys.readouterr()
        out = out.splitlines()
        assert len(out) == 4
        assert out[2].endswith("test0.fits 49491.65366175    0.23")
        assert out[3].endswith("test1.fits 49492.65366175    0.22")

        fitsheader.main(
            [
                "-e",
                "0",
                "-f",
                "-k",
                "EXPSTART",
                "-k",
                "EXPTIME",
                self.data("test0.fits"),
            ]
        )
        out, err = capsys.readouterr()
        out = out.splitlines()
        assert len(out) == 3
        assert out[2].endswith("test0.fits 49491.65366175    0.23")

        fitsheader.main(
            ["-f", "-k", "NAXIS", self.data("tdim.fits"), self.data("test1.fits")]
        )
        out, err = capsys.readouterr()
        out = out.splitlines()
        assert len(out) == 4
        assert out[0].endswith("0:NAXIS 1:NAXIS 2:NAXIS 3:NAXIS 4:NAXIS")
        assert out[2].endswith("tdim.fits       0       2      --      --      --")
        assert out[3].endswith("test1.fits       0       2       2       2       2")

        # check that files without required keyword are present
        fitsheader.main(
            ["-f", "-k", "DATE-OBS", self.data("table.fits"), self.data("test0.fits")]
        )
        out, err = capsys.readouterr()
        out = out.splitlines()
        assert len(out) == 4
        assert out[2].endswith("table.fits       --")
        assert out[3].endswith("test0.fits 19/05/94")

        # check that COMMENT and HISTORY are excluded
        fitsheader.main(["-e", "0", "-f", self.data("tb.fits")])
        out, err = capsys.readouterr()
        out = out.splitlines()
        assert len(out) == 3
        assert out[2].endswith(
            "tb.fits   True     16     0   True STScI-STSDAS/TABLES  tb.fits       1"
        )

    def test_fitsort_sorting_keyword_fitsort(self, capsys):
        """check that sorting by keyword works"""
        fitsheader.main(
            [
                "-f",
                "-k",
                "NAXIS",
                "-e",
                "0",
                self.data("group.fits"),
                self.data("test0.fits"),
            ]
        )
        out_unsorted, err_unsorted = capsys.readouterr()
        out_unsorted = out_unsorted.splitlines()

        fitsheader.main(
            [
                "-f",
                "-s",
                "NAXIS",
                "-k",
                "NAXIS",
                "-e",
                "0",
                self.data("group.fits"),
                self.data("test0.fits"),
            ]
        )
        out_sorted, err_sorted = capsys.readouterr()
        out_sorted = out_sorted.splitlines()

        assert len(out_unsorted) == 4
        assert out_unsorted[2].endswith("group.fits     5")
        assert out_unsorted[3].endswith("test0.fits     0")

        assert len(out_sorted) == 4
        assert out_sorted[2].endswith("test0.fits     0")
        assert out_sorted[3].endswith("group.fits     5")

    def test_fitsort_sorting_keyword_complains(self, capsys):
        with pytest.raises(SystemExit):
            fitsheader.main(
                ["-t", "-s", "DUMMY", self.data("group.fits"), self.data("test0.fits")]
            )
        out_table, err_table = capsys.readouterr()
        assert "only supported in conjunction with -f/--fitsort" in err_table
        with pytest.raises(SystemExit):
            fitsheader.main(
                ["-s", "DUMMY", self.data("group.fits"), self.data("test0.fits")]
            )
        out_default, err_default = capsys.readouterr()
        assert "only supported in conjunction with -f/--fitsort" in err_default

    def test_dotkeyword(self, capsys):
        fitsheader.main(["-e", "0", "-k", "ESO DET ID", self.data("fixed-1890.fits")])
        out, err = capsys.readouterr()
        out = out.splitlines()
        assert len(out) == 2
        assert (
            out[1].strip().endswith("HIERARCH ESO DET ID = 'DV13' / Detector system Id")
        )

        fitsheader.main(["-e", "0", "-k", "ESO.DET.ID", self.data("fixed-1890.fits")])
        out, err = capsys.readouterr()
        out = out.splitlines()
        assert len(out) == 2
        assert (
            out[1].strip().endswith("HIERARCH ESO DET ID = 'DV13' / Detector system Id")
        )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
``fitscheck`` is a command line script based on astropy.io.fits for verifying
and updating the CHECKSUM and DATASUM keywords of .fits files.  ``fitscheck``
can also detect and often fix other FITS standards violations.  ``fitscheck``
facilitates re-writing the non-standard checksums originally generated by
astropy.io.fits with standard checksums which will interoperate with CFITSIO.

``fitscheck`` will refuse to write new checksums if the checksum keywords are
missing or their values are bad.  Use ``--force`` to write new checksums
regardless of whether or not they currently exist or pass.  Use
``--ignore-missing`` to tolerate missing checksum keywords without comment.

Example uses of fitscheck:

1. Add checksums::

    $ fitscheck --write *.fits

2. Write new checksums, even if existing checksums are bad or missing::

    $ fitscheck --write --force *.fits

3. Verify standard checksums and FITS compliance without changing the files::

    $ fitscheck --compliance *.fits

4. Only check and fix compliance problems,  ignoring checksums::

    $ fitscheck --checksum none --compliance --write *.fits

5. Verify standard interoperable checksums::

    $ fitscheck *.fits

6. Delete checksum keywords::

    $ fitscheck --checksum remove --write *.fits

"""

import argparse
import logging
import sys
import warnings

from astropy import __version__
from astropy.io import fits

log = logging.getLogger("fitscheck")

DESCRIPTION = """
e.g. fitscheck example.fits

Verifies and optionally re-writes the CHECKSUM and DATASUM keywords
for a .fits file.
Optionally detects and fixes FITS standard compliance problems.

This script is part of the Astropy package. See
https://docs.astropy.org/en/latest/io/fits/usage/scripts.html#module-astropy.io.fits.scripts.fitscheck
for further documentation.
""".strip()


def handle_options(args):
    if not len(args):
        args = ["-h"]

    parser = argparse.ArgumentParser(
        description=DESCRIPTION, formatter_class=argparse.RawDescriptionHelpFormatter
    )

    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )

    parser.add_argument(
        "fits_files", metavar="file", nargs="+", help=".fits files to process."
    )

    parser.add_argument(
        "-k",
        "--checksum",
        dest="checksum_kind",
        choices=["standard", "remove", "none"],
        help="Choose FITS checksum mode or none.  Defaults standard.",
        default="standard",
    )

    parser.add_argument(
        "-w",
        "--write",
        dest="write_file",
        help="Write out file checksums and/or FITS compliance fixes.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "-f",
        "--force",
        dest="force",
        help="Do file update even if original checksum was bad.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "-c",
        "--compliance",
        dest="compliance",
        help="Do FITS compliance checking; fix if possible.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "-i",
        "--ignore-missing",
        dest="ignore_missing",
        help="Ignore missing checksums.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        help="Generate extra output.",
        default=False,
        action="store_true",
    )

    global OPTIONS
    OPTIONS = parser.parse_args(args)

    if OPTIONS.checksum_kind == "none":
        OPTIONS.checksum_kind = False
    elif OPTIONS.checksum_kind == "standard":
        OPTIONS.checksum_kind = True
    elif OPTIONS.checksum_kind == "remove":
        OPTIONS.write_file = True
        OPTIONS.force = True

    return OPTIONS.fits_files


def setup_logging():
    log.handlers.clear()

    if OPTIONS.verbose:
        log.setLevel(logging.INFO)
    else:
        log.setLevel(logging.WARNING)

    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("%(message)s"))
    log.addHandler(handler)


def verify_checksums(filename):
    """
    Prints a message if any HDU in `filename` has a bad checksum or datasum.
    """
    with warnings.catch_warnings(record=True) as wlist:
        warnings.simplefilter("always")
        with fits.open(filename, checksum=OPTIONS.checksum_kind) as hdulist:
            for i, hdu in enumerate(hdulist):
                # looping on HDUs is needed to read them and verify the
                # checksums
                if not OPTIONS.ignore_missing:
                    if not hdu._checksum:
                        log.warning(
                            f"MISSING {filename!r} .. Checksum not found in HDU #{i}"
                        )
                        return 1
                    if not hdu._datasum:
                        log.warning(
                            f"MISSING {filename!r} .. Datasum not found in HDU #{i}"
                        )
                        return 1

    for w in wlist:
        if str(w.message).startswith(
            ("Checksum verification failed", "Datasum verification failed")
        ):
            log.warning("BAD %r %s", filename, str(w.message))
            return 1

    log.info(f"OK {filename!r}")
    return 0


def verify_compliance(filename):
    """Check for FITS standard compliance."""
    with fits.open(filename) as hdulist:
        try:
            hdulist.verify("exception")
        except fits.VerifyError as exc:
            log.warning("NONCOMPLIANT %r .. %s", filename, str(exc).replace("\n", " "))
            return 1
    return 0


def update(filename):
    """
    Sets the ``CHECKSUM`` and ``DATASUM`` keywords for each HDU of `filename`.

    Also updates fixes standards violations if possible and requested.
    """
    output_verify = "silentfix" if OPTIONS.compliance else "ignore"

    # For unit tests we reset temporarily the warning filters. Indeed, before
    # updating the checksums, fits.open will verify the existing checksums and
    # raise warnings, which are later caught and converted to log.warning...
    # which is an issue when testing, using the "error" action to convert
    # warnings to exceptions.
    with warnings.catch_warnings():
        warnings.resetwarnings()
        with fits.open(
            filename,
            do_not_scale_image_data=True,
            checksum=OPTIONS.checksum_kind,
            mode="update",
        ) as hdulist:
            hdulist.flush(output_verify=output_verify)


def process_file(filename):
    """
    Handle a single .fits file,  returning the count of checksum and compliance
    errors.
    """
    try:
        checksum_errors = verify_checksums(filename)
        if OPTIONS.compliance:
            compliance_errors = verify_compliance(filename)
        else:
            compliance_errors = 0
        if OPTIONS.write_file and checksum_errors == 0 or OPTIONS.force:
            update(filename)
        return checksum_errors + compliance_errors
    except Exception as e:
        log.error(f"EXCEPTION {filename!r} .. {e}")
        return 1


def main(args=None):
    """
    Processes command line parameters into options and files,  then checks
    or update FITS DATASUM and CHECKSUM keywords for the specified files.
    """
    errors = 0
    fits_files = handle_options(args or sys.argv[1:])
    setup_logging()
    for filename in fits_files:
        errors += process_file(filename)
    if errors:
        log.warning(f"{errors} errors")
    return int(bool(errors))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import functools

import numpy as np
import pytest

from astropy.time import Time
from astropy.utils.iers import conf as iers_conf
from astropy.utils.iers import iers  # used in testing

allclose_jd = functools.partial(np.allclose, rtol=0, atol=1e-9)
allclose_sec = functools.partial(np.allclose, rtol=1e-15, atol=1e-4)
# 0.1 ms atol; IERS-B files change at that level.

try:
    iers.IERS_A.open()  # check if IERS_A is available
except OSError:
    HAS_IERS_A = False
else:
    HAS_IERS_A = True


def do_ut1_prediction_tst(iers_type):
    tnow = Time.now()
    iers_tab = iers_type.open()
    tnow.delta_ut1_utc, status = iers_tab.ut1_utc(tnow, return_status=True)
    assert status == iers.FROM_IERS_A_PREDICTION
    tnow_ut1_jd = tnow.ut1.jd
    assert allclose_jd(tnow_ut1_jd - tnow.jd, tnow.delta_ut1_utc / 86400)

    delta_ut1_utc = tnow.delta_ut1_utc
    with iers.earth_orientation_table.set(iers_type.open()):
        delta2, status2 = tnow.get_delta_ut1_utc(return_status=True)
        assert status2 == status
        assert delta2.to_value("s") == delta_ut1_utc

        tnow_ut1 = tnow.ut1
        assert tnow_ut1._delta_ut1_utc == delta_ut1_utc
        assert allclose_jd(tnow_ut1.jd - tnow.jd, tnow.delta_ut1_utc / 86400)


@pytest.mark.remote_data
class TestTimeUT1Remote:
    def setup_class(cls):
        # Need auto_download so that IERS_B won't be loaded and cause tests to
        # fail.
        iers_conf.auto_download = True

    def teardown_class(cls):
        # This setting is to be consistent with astropy/conftest.py
        iers_conf.auto_download = False

    def test_utc_to_ut1(self):
        "Test conversion of UTC to UT1, making sure to include a leap second"
        t = Time(
            [
                "2012-06-30 12:00:00",
                "2012-06-30 23:59:59",
                "2012-06-30 23:59:60",
                "2012-07-01 00:00:00",
                "2012-07-01 12:00:00",
            ],
            scale="utc",
        )
        t_ut1_jd = t.ut1.jd
        t_comp = np.array(
            [
                2456108.9999932079,
                2456109.4999816339,
                2456109.4999932083,
                2456109.5000047823,
                2456110.0000047833,
            ]
        )
        assert allclose_jd(t_ut1_jd, t_comp)
        t_back = t.ut1.utc
        assert allclose_jd(t.jd, t_back.jd)

        tnow = Time.now()

        tnow.ut1

    def test_ut1_iers_auto(self):
        do_ut1_prediction_tst(iers.IERS_Auto)


class TestTimeUT1:
    """Test Time.ut1 using IERS tables"""

    def test_ut1_to_utc(self):
        """Also test the reverse, around the leap second
        (round-trip test closes #2077)"""
        with iers_conf.set_temp("auto_download", False):
            t = Time(
                [
                    "2012-06-30 12:00:00",
                    "2012-06-30 23:59:59",
                    "2012-07-01 00:00:00",
                    "2012-07-01 00:00:01",
                    "2012-07-01 12:00:00",
                ],
                scale="ut1",
            )
            t_utc_jd = t.utc.jd
            t_comp = np.array(
                [
                    2456109.0000010049,
                    2456109.4999836441,
                    2456109.4999952177,
                    2456109.5000067917,
                    2456109.9999952167,
                ]
            )
            assert allclose_jd(t_utc_jd, t_comp)
            t_back = t.utc.ut1
            assert allclose_jd(t.jd, t_back.jd)

    def test_empty_ut1(self):
        """Testing for a zero-length Time object from UTC to UT1
        when an empty array is passed"""
        from astropy import units as u

        with iers_conf.set_temp("auto_download", False):
            t = Time(["2012-06-30 12:00:00"]) + np.arange(24) * u.hour
            t_empty = t[[]].ut1
            assert isinstance(t_empty, Time)
            assert t_empty.scale == "ut1"
            assert t_empty.size == 0

    def test_delta_ut1_utc(self):
        """Accessing delta_ut1_utc should try to get it from IERS
        (closes #1924 partially)"""
        with iers_conf.set_temp("auto_download", False):
            t = Time("2012-06-30 12:00:00", scale="utc")
            assert not hasattr(t, "_delta_ut1_utc")
            # accessing delta_ut1_utc calculates it
            assert allclose_sec(t.delta_ut1_utc, -0.58682110003124965)
            # and keeps it around
            assert allclose_sec(t._delta_ut1_utc, -0.58682110003124965)


class TestTimeUT1SpecificIERSTable:
    @pytest.mark.skipif(not HAS_IERS_A, reason="requires IERS_A")
    def test_ut1_iers_A(self):
        do_ut1_prediction_tst(iers.IERS_A)

    def test_ut1_iers_B(self):
        tnow = Time.now()
        iers_b = iers.IERS_B.open()
        delta1, status1 = tnow.get_delta_ut1_utc(iers_b, return_status=True)
        assert status1 == iers.TIME_BEYOND_IERS_RANGE

        with iers.earth_orientation_table.set(iers.IERS_B.open()):
            delta2, status2 = tnow.get_delta_ut1_utc(return_status=True)
            assert status2 == status1

            with pytest.raises(iers.IERSRangeError):
                tnow.ut1



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import annotations

__all__: list[str] = []  # nothing is publicly scoped

import sys
from typing import TYPE_CHECKING

from astropy.utils.decorators import deprecated

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any


# TODO: complete the deprecation for v6.2 / v7
def __getattr__(name: str) -> Callable[..., Any]:
    """Get realizations using lazy import from ``PEP 562``.

    Raises
    ------
    AttributeError
        If "name" is not in :mod:`astropy.cosmology.realizations`
    """
    if name not in ("vectorize_redshift_method", "aszarr"):
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}.")

    from . import _utils

    func = deprecated(
        since="v6.0",
        message=(
            "this private function has been moved to the private module"
            " `astropy.cosmology._utils`"
        ),
    )(getattr(_utils, name))

    sys.modules[__name__].__dict__[name] = func  # cache for next time

    return func



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Read/Write/Interchange methods for `astropy.cosmology`.

**NOT public API**.
"""

# Import to register with the I/O machinery
from . import cosmology, ecsv, html, latex, mapping, model, row, table, yaml



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import operator

__all__ = ["BST"]


class MaxValue:
    """
    Represents an infinite value for purposes
    of tuple comparison.
    """

    def __gt__(self, other):
        return True

    def __ge__(self, other):
        return True

    def __lt__(self, other):
        return False

    def __le__(self, other):
        return False

    def __repr__(self):
        return "MAX"

    __str__ = __repr__


class MinValue:
    """
    The opposite of MaxValue, i.e. a representation of
    negative infinity.
    """

    def __lt__(self, other):
        return True

    def __le__(self, other):
        return True

    def __gt__(self, other):
        return False

    def __ge__(self, other):
        return False

    def __repr__(self):
        return "MIN"

    __str__ = __repr__


class Epsilon:
    """
    Represents the "next largest" version of a given value,
    so that for all valid comparisons we have
    x < y < Epsilon(y) < z whenever x < y < z and x, z are
    not Epsilon objects.

    Parameters
    ----------
    val : object
        Original value
    """

    __slots__ = ("val",)

    def __init__(self, val):
        self.val = val

    def __lt__(self, other):
        if self.val == other:
            return False
        return self.val < other

    def __gt__(self, other):
        if self.val == other:
            return True
        return self.val > other

    def __eq__(self, other):
        return False

    def __repr__(self):
        return repr(self.val) + " + epsilon"


class Node:
    """
    An element in a binary search tree, containing
    a key, data, and references to children nodes and
    a parent node.

    Parameters
    ----------
    key : tuple
        Node key
    data : list or int
        Node data
    """

    __lt__ = lambda x, y: x.key < y.key
    __le__ = lambda x, y: x.key <= y.key
    __eq__ = lambda x, y: x.key == y.key
    __ge__ = lambda x, y: x.key >= y.key
    __gt__ = lambda x, y: x.key > y.key
    __ne__ = lambda x, y: x.key != y.key
    __slots__ = ("key", "data", "left", "right")

    # each node has a key and data list
    def __init__(self, key, data):
        self.key = key
        self.data = data if isinstance(data, list) else [data]
        self.left = None
        self.right = None

    def replace(self, child, new_child):
        """
        Replace this node's child with a new child.
        """
        if self.left is not None and self.left == child:
            self.left = new_child
        elif self.right is not None and self.right == child:
            self.right = new_child
        else:
            raise ValueError("Cannot call replace() on non-child")

    def remove(self, child):
        """
        Remove the given child.
        """
        self.replace(child, None)

    def set(self, other):
        """
        Copy the given node.
        """
        self.key = other.key
        self.data = other.data[:]

    def __str__(self):
        return str((self.key, self.data))

    def __repr__(self):
        return str(self)


class BST:
    """
    A basic binary search tree in pure Python, used
    as an engine for indexing.

    Parameters
    ----------
    data : Table
        Sorted columns of the original table
    row_index : Column object
        Row numbers corresponding to data columns
    unique : bool
        Whether the values of the index must be unique.
        Defaults to False.
    """

    NodeClass = Node

    def __init__(self, data, row_index, unique=False):
        self.root = None
        self.size = 0
        self.unique = unique
        for key, row in zip(data, row_index):
            self.add(tuple(key), row)

    def add(self, key, data=None):
        """
        Add a key, data pair.
        """
        if data is None:
            data = key

        self.size += 1
        node = self.NodeClass(key, data)
        curr_node = self.root
        if curr_node is None:
            self.root = node
            return
        while True:
            if node < curr_node:
                if curr_node.left is None:
                    curr_node.left = node
                    break
                curr_node = curr_node.left
            elif node > curr_node:
                if curr_node.right is None:
                    curr_node.right = node
                    break
                curr_node = curr_node.right
            elif self.unique:
                raise ValueError("Cannot insert non-unique value")
            else:  # add data to node
                curr_node.data.extend(node.data)
                curr_node.data = sorted(curr_node.data)
                return

    def find(self, key):
        """
        Return all data values corresponding to a given key.

        Parameters
        ----------
        key : tuple
            Input key

        Returns
        -------
        data_vals : list
            List of rows corresponding to the input key
        """
        node, parent = self.find_node(key)
        return node.data if node is not None else []

    def find_node(self, key):
        """
        Find the node associated with the given key.
        """
        if self.root is None:
            return (None, None)
        return self._find_recursive(key, self.root, None)

    def shift_left(self, row):
        """
        Decrement all rows larger than the given row.
        """
        for node in self.traverse():
            node.data = [x - 1 if x > row else x for x in node.data]

    def shift_right(self, row):
        """
        Increment all rows greater than or equal to the given row.
        """
        for node in self.traverse():
            node.data = [x + 1 if x >= row else x for x in node.data]

    def _find_recursive(self, key, node, parent):
        try:
            if key == node.key:
                return (node, parent)
            elif key > node.key:
                if node.right is None:
                    return (None, None)
                return self._find_recursive(key, node.right, node)
            else:
                if node.left is None:
                    return (None, None)
                return self._find_recursive(key, node.left, node)
        except TypeError:  # wrong key type
            return (None, None)

    def traverse(self, order="inorder"):
        """
        Return nodes of the BST in the given order.

        Parameters
        ----------
        order : str
            The order in which to recursively search the BST.
            Possible values are:
            "preorder": current node, left subtree, right subtree
            "inorder": left subtree, current node, right subtree
            "postorder": left subtree, right subtree, current node
        """
        if order == "preorder":
            return self._preorder(self.root, [])
        elif order == "inorder":
            return self._inorder(self.root, [])
        elif order == "postorder":
            return self._postorder(self.root, [])
        raise ValueError(f'Invalid traversal method: "{order}"')

    def items(self):
        """
        Return BST items in order as (key, data) pairs.
        """
        return [(x.key, x.data) for x in self.traverse()]

    def sort(self):
        """
        Make row order align with key order.
        """
        i = 0
        for node in self.traverse():
            num_rows = len(node.data)
            node.data = list(range(i, i + num_rows))
            i += num_rows

    def sorted_data(self):
        """
        Return BST rows sorted by key values.
        """
        return [x for node in self.traverse() for x in node.data]

    def _preorder(self, node, lst):
        if node is None:
            return lst
        lst.append(node)
        self._preorder(node.left, lst)
        self._preorder(node.right, lst)
        return lst

    def _inorder(self, node, lst):
        if node is None:
            return lst
        self._inorder(node.left, lst)
        lst.append(node)
        self._inorder(node.right, lst)
        return lst

    def _postorder(self, node, lst):
        if node is None:
            return lst
        self._postorder(node.left, lst)
        self._postorder(node.right, lst)
        lst.append(node)
        return lst

    def _substitute(self, node, parent, new_node):
        if node is self.root:
            self.root = new_node
        else:
            parent.replace(node, new_node)

    def remove(self, key, data=None):
        """
        Remove data corresponding to the given key.

        Parameters
        ----------
        key : tuple
            The key to remove
        data : int or None
            If None, remove the node corresponding to the given key.
            If not None, remove only the given data value from the node.

        Returns
        -------
        successful : bool
            True if removal was successful, false otherwise
        """
        node, parent = self.find_node(key)
        if node is None:
            return False
        if data is not None:
            if data not in node.data:
                raise ValueError("Data does not belong to correct node")
            elif len(node.data) > 1:
                node.data.remove(data)
                return True
        if node.left is None and node.right is None:
            self._substitute(node, parent, None)
        elif node.left is None and node.right is not None:
            self._substitute(node, parent, node.right)
        elif node.right is None and node.left is not None:
            self._substitute(node, parent, node.left)
        else:
            # find largest element of left subtree
            curr_node = node.left
            parent = node
            while curr_node.right is not None:
                parent = curr_node
                curr_node = curr_node.right
            self._substitute(curr_node, parent, curr_node.left)
            node.set(curr_node)
        self.size -= 1
        return True

    def is_valid(self):
        """
        Returns whether this is a valid BST.
        """
        return self._is_valid(self.root)

    def _is_valid(self, node):
        if node is None:
            return True
        return (
            (node.left is None or node.left <= node)
            and (node.right is None or node.right >= node)
            and self._is_valid(node.left)
            and self._is_valid(node.right)
        )

    def range(self, lower, upper, bounds=(True, True)):
        """
        Return all nodes with keys in the given range.

        Parameters
        ----------
        lower : tuple
            Lower bound
        upper : tuple
            Upper bound
        bounds : (2,) tuple of bool
            Indicates whether the search should be inclusive or
            exclusive with respect to the endpoints. The first
            argument corresponds to an inclusive lower bound,
            and the second argument to an inclusive upper bound.
        """
        nodes = self.range_nodes(lower, upper, bounds)
        return [x for node in nodes for x in node.data]

    def range_nodes(self, lower, upper, bounds=(True, True)):
        """
        Return nodes in the given range.
        """
        if self.root is None:
            return []
        # op1 is <= or <, op2 is >= or >
        op1 = operator.le if bounds[0] else operator.lt
        op2 = operator.ge if bounds[1] else operator.gt
        return self._range(lower, upper, op1, op2, self.root, [])

    def same_prefix(self, val):
        """
        Assuming the given value has smaller length than keys, return
        nodes whose keys have this value as a prefix.
        """
        if self.root is None:
            return []
        nodes = self._same_prefix(val, self.root, [])
        return [x for node in nodes for x in node.data]

    def _range(self, lower, upper, op1, op2, node, lst):
        if op1(lower, node.key) and op2(upper, node.key):
            lst.append(node)
        if upper > node.key and node.right is not None:
            self._range(lower, upper, op1, op2, node.right, lst)
        if lower < node.key and node.left is not None:
            self._range(lower, upper, op1, op2, node.left, lst)
        return lst

    def _same_prefix(self, val, node, lst):
        prefix = node.key[: len(val)]
        if prefix == val:
            lst.append(node)
        if prefix <= val and node.right is not None:
            self._same_prefix(val, node.right, lst)
        if prefix >= val and node.left is not None:
            self._same_prefix(val, node.left, lst)
        return lst

    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    def _print(self, node, level):
        line = "\t" * level + str(node) + "\n"
        if node.left is not None:
            line += self._print(node.left, level + 1)
        if node.right is not None:
            line += self._print(node.right, level + 1)
        return line

    @property
    def height(self):
        """
        Return the BST height.
        """
        return self._height(self.root)

    def _height(self, node):
        if node is None:
            return -1
        return max(self._height(node.left), self._height(node.right)) + 1

    def replace_rows(self, row_map):
        """
        Replace all rows with the values they map to in the
        given dictionary. Any rows not present as keys in
        the dictionary will have their nodes deleted.

        Parameters
        ----------
        row_map : dict
            Mapping of row numbers to new row numbers
        """
        for _, data in self.items():  # noqa: PERF102
            data[:] = [row_map[x] for x in data if x in row_map]



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This subpackage contains classes and functions for defining and converting
between different physical units.

This code is adapted from the `pynbody
<https://github.com/pynbody/pynbody>`_ units module written by Andrew
Pontzen, who has granted the Astropy project permission to use the
code under a BSD license.
"""

from . import (
    astrophys,
    cgs,
    core,
    decorators,
    misc,
    photometric,
    physical,
    quantity,
    si,
    structured,
)
from .astrophys import *
from .cgs import *
from .core import *
from .core import set_enabled_units
from .decorators import *
from .misc import *
from .photometric import *
from .physical import *
from .quantity import *
from .si import *
from .structured import *

# Import order matters here -- circular dependencies abound!
# isort: split
from . import equivalencies, function
from .equivalencies import *
from .function import *

__all__ = []
__all__ += core.__all__
__all__ += quantity.__all__
__all__ += decorators.__all__
__all__ += structured.__all__
__all__ += si.__all__
__all__ += cgs.__all__
__all__ += astrophys.__all__
__all__ += physical.__all__
__all__ += misc.__all__
__all__ += photometric.__all__
__all__ += structured.__all__
__all__ += equivalencies.__all__
__all__ += function.__all__
# We sort `__all__` for the docs. We use `globals` to avoid confusing any static
# analysis tools.
globals()["__all__"].sort()

# Enable the set of default units.  This notably does *not* include
# Imperial units.
set_enabled_units([si, cgs, astrophys, function.units, misc, photometric])



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

# This file was automatically generated from ply. To re-generate this file,
# remove it from this folder, then build astropy and run the tests in-place:
#
#   python setup.py build_ext --inplace
#   pytest astropy/units
#
# You can then commit the changes to this file.


# cds_parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'CLOSE_BRACKET CLOSE_PAREN DIMENSIONLESS DIVISION OPEN_BRACKET OPEN_PAREN PRODUCT SIGN UFLOAT UINT UNIT X\n            main : factor combined_units\n                 | combined_units\n                 | DIMENSIONLESS\n                 | OPEN_BRACKET combined_units CLOSE_BRACKET\n                 | OPEN_BRACKET DIMENSIONLESS CLOSE_BRACKET\n                 | factor\n            \n            combined_units : product_of_units\n                           | division_of_units\n            \n            product_of_units : unit_expression PRODUCT combined_units\n                             | unit_expression\n            \n            division_of_units : DIVISION unit_expression\n                              | combined_units DIVISION unit_expression\n            \n            unit_expression : unit_with_power\n                            | OPEN_PAREN combined_units CLOSE_PAREN\n            \n            factor : signed_float X UINT signed_int\n                   | UINT X UINT signed_int\n                   | UINT signed_int\n                   | UINT\n                   | signed_float\n            \n            unit_with_power : UNIT numeric_power\n                            | UNIT\n            \n            numeric_power : sign UINT\n            \n            sign : SIGN\n                 |\n            \n            signed_int : SIGN UINT\n            \n            signed_float : sign UINT\n                         | sign UFLOAT\n            '
    
_lr_action_items = {'DIMENSIONLESS':([0,5,],[4,20,]),'OPEN_BRACKET':([0,],[5,]),'UINT':([0,10,13,16,21,22,24,31,],[7,25,-23,-24,35,36,37,40,]),'DIVISION':([0,2,3,5,6,7,8,9,11,14,15,16,17,19,23,25,26,27,28,29,30,32,37,38,39,40,41,42,],[12,12,18,12,-19,-18,-7,-8,-10,-13,12,-21,18,18,-17,-26,-27,12,-11,18,-20,-12,-25,18,-14,-22,-15,-16,]),'SIGN':([0,7,16,35,36,],[13,24,13,24,24,]),'UFLOAT':([0,10,13,],[-24,26,-23,]),'OPEN_PAREN':([0,2,5,6,7,12,15,18,23,25,26,27,37,41,42,],[15,15,15,-19,-18,15,15,15,-17,-26,-27,15,-25,-15,-16,]),'UNIT':([0,2,5,6,7,12,15,18,23,25,26,27,37,41,42,],[16,16,16,-19,-18,16,16,16,-17,-26,-27,16,-25,-15,-16,]),'$end':([1,2,3,4,6,7,8,9,11,14,16,17,23,25,26,28,30,32,33,34,37,38,39,40,41,42,],[0,-6,-2,-3,-19,-18,-7,-8,-10,-13,-21,-1,-17,-26,-27,-11,-20,-12,-4,-5,-25,-9,-14,-22,-15,-16,]),'X':([6,7,25,26,],[21,22,-26,-27,]),'CLOSE_BRACKET':([8,9,11,14,16,19,20,28,30,32,38,39,40,],[-7,-8,-10,-13,-21,33,34,-11,-20,-12,-9,-14,-22,]),'CLOSE_PAREN':([8,9,11,14,16,28,29,30,32,38,39,40,],[-7,-8,-10,-13,-21,-11,39,-20,-12,-9,-14,-22,]),'PRODUCT':([11,14,16,30,39,40,],[27,-13,-21,-20,-14,-22,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'main':([0,],[1,]),'factor':([0,],[2,]),'combined_units':([0,2,5,15,27,],[3,17,19,29,38,]),'signed_float':([0,],[6,]),'product_of_units':([0,2,5,15,27,],[8,8,8,8,8,]),'division_of_units':([0,2,5,15,27,],[9,9,9,9,9,]),'sign':([0,16,],[10,31,]),'unit_expression':([0,2,5,12,15,18,27,],[11,11,11,28,11,32,11,]),'unit_with_power':([0,2,5,12,15,18,27,],[14,14,14,14,14,14,14,]),'signed_int':([7,35,36,],[23,41,42,]),'numeric_power':([16,],[30,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> main","S'",1,None,None,None),
  ('main -> factor combined_units','main',2,'p_main','cds.py',143),
  ('main -> combined_units','main',1,'p_main','cds.py',144),
  ('main -> DIMENSIONLESS','main',1,'p_main','cds.py',145),
  ('main -> OPEN_BRACKET combined_units CLOSE_BRACKET','main',3,'p_main','cds.py',146),
  ('main -> OPEN_BRACKET DIMENSIONLESS CLOSE_BRACKET','main',3,'p_main','cds.py',147),
  ('main -> factor','main',1,'p_main','cds.py',148),
  ('combined_units -> product_of_units','combined_units',1,'p_combined_units','cds.py',162),
  ('combined_units -> division_of_units','combined_units',1,'p_combined_units','cds.py',163),
  ('product_of_units -> unit_expression PRODUCT combined_units','product_of_units',3,'p_product_of_units','cds.py',169),
  ('product_of_units -> unit_expression','product_of_units',1,'p_product_of_units','cds.py',170),
  ('division_of_units -> DIVISION unit_expression','division_of_units',2,'p_division_of_units','cds.py',179),
  ('division_of_units -> combined_units DIVISION unit_expression','division_of_units',3,'p_division_of_units','cds.py',180),
  ('unit_expression -> unit_with_power','unit_expression',1,'p_unit_expression','cds.py',189),
  ('unit_expression -> OPEN_PAREN combined_units CLOSE_PAREN','unit_expression',3,'p_unit_expression','cds.py',190),
  ('factor -> signed_float X UINT signed_int','factor',4,'p_factor','cds.py',199),
  ('factor -> UINT X UINT signed_int','factor',4,'p_factor','cds.py',200),
  ('factor -> UINT signed_int','factor',2,'p_factor','cds.py',201),
  ('factor -> UINT','factor',1,'p_factor','cds.py',202),
  ('factor -> signed_float','factor',1,'p_factor','cds.py',203),
  ('unit_with_power -> UNIT numeric_power','unit_with_power',2,'p_unit_with_power','cds.py',218),
  ('unit_with_power -> UNIT','unit_with_power',1,'p_unit_with_power','cds.py',219),
  ('numeric_power -> sign UINT','numeric_power',2,'p_numeric_power','cds.py',228),
  ('sign -> SIGN','sign',1,'p_sign','cds.py',234),
  ('sign -> <empty>','sign',0,'p_sign','cds.py',235),
  ('signed_int -> SIGN UINT','signed_int',2,'p_signed_int','cds.py',244),
  ('signed_float -> sign UINT','signed_float',2,'p_signed_float','cds.py',250),
  ('signed_float -> sign UFLOAT','signed_float',2,'p_signed_float','cds.py',251),
]



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np

from astropy.io.fits.hdu.base import BITPIX2DTYPE
from astropy.io.fits.hdu.compressed._tiled_compression import (
    decompress_image_data_section,
)
from astropy.io.fits.hdu.compressed.utils import _data_shape, _n_tiles, _tile_shape
from astropy.utils.shapes import simplify_basic_index


class CompImageSection:
    """
    Class enabling subsets of CompImageHDU data to be loaded lazily via slicing.

    Slices of this object load the corresponding section of an image array from
    the underlying FITS file, and applies any BSCALE/BZERO factors.

    Section slices cannot be assigned to, and modifications to a section are
    not saved back to the underlying file.

    See the :ref:`astropy:data-sections` section of the Astropy documentation
    for more details.
    """

    def __init__(self, hdu):
        self.hdu = hdu
        self._data_shape = _data_shape(self.hdu._header)
        self._tile_shape = _tile_shape(self.hdu._header)
        self._n_dim = len(self._data_shape)
        self._n_tiles = np.array(
            _n_tiles(self._data_shape, self._tile_shape), dtype=int
        )

    @property
    def shape(self):
        return tuple(self._data_shape)

    @property
    def ndim(self):
        return self.hdu._header["ZNAXIS"]

    @property
    def dtype(self):
        return BITPIX2DTYPE[self.hdu._header["ZBITPIX"]]

    def __getitem__(self, index):
        # Shortcut if the whole data is requested (this is used by the
        # data property, so we optimize it as it is frequently used)
        if index is Ellipsis:
            first_tile_index = np.zeros(self._n_dim, dtype=int)
            last_tile_index = self._n_tiles - 1
            data = decompress_image_data_section(
                self.hdu.compressed_data,
                self.hdu.compression_type,
                self.hdu._header,
                self.hdu,
                first_tile_index,
                last_tile_index,
            )
            return self.hdu._scale_data(data)

        index = simplify_basic_index(index, shape=self._data_shape)

        # Determine for each dimension the first and last tile to extract

        first_tile_index = np.zeros(self._n_dim, dtype=int)
        last_tile_index = np.zeros(self._n_dim, dtype=int)

        final_array_index = []

        for dim, idx in enumerate(index):
            if isinstance(idx, slice):
                if idx.step > 0:
                    first_tile_index[dim] = idx.start // self._tile_shape[dim]
                    last_tile_index[dim] = (idx.stop - 1) // self._tile_shape[dim]
                else:
                    stop = 0 if idx.stop is None else max(idx.stop - 1, 0)
                    first_tile_index[dim] = stop // self._tile_shape[dim]
                    last_tile_index[dim] = idx.start // self._tile_shape[dim]

                # Because slices such as slice(5, 0, 1) can exist (which
                # would be empty) we need to make sure last_tile_index is
                # always larger than first_tile_index
                last_tile_index = np.maximum(last_tile_index, first_tile_index)

                if idx.step < 0 and idx.stop is None:
                    final_array_index.append(idx)
                else:
                    final_array_index.append(
                        slice(
                            idx.start - self._tile_shape[dim] * first_tile_index[dim],
                            idx.stop - self._tile_shape[dim] * first_tile_index[dim],
                            idx.step,
                        )
                    )
            else:
                first_tile_index[dim] = idx // self._tile_shape[dim]
                last_tile_index[dim] = first_tile_index[dim]
                final_array_index.append(
                    idx - self._tile_shape[dim] * first_tile_index[dim]
                )

        data = decompress_image_data_section(
            self.hdu.compressed_data,
            self.hdu.compression_type,
            self.hdu._header,
            self.hdu,
            first_tile_index,
            last_tile_index,
        )

        return self.hdu._scale_data(data[tuple(final_array_index)])



# Licensed under a 3-clause BSD style license - see PYFITS.rst

import sys
import warnings

import numpy as np
import pytest

from astropy.io import fits
from astropy.io.fits.hdu.base import _ValidHDU

from .conftest import FitsTestCase
from .test_table import comparerecords


class BaseChecksumTests(FitsTestCase):
    def setup_method(self):
        super().setup_method()
        self._oldfilters = warnings.filters[:]
        warnings.filterwarnings("error", message="Checksum verification failed")
        warnings.filterwarnings("error", message="Datasum verification failed")

        # Monkey-patch the _get_timestamp method so that the checksum
        # timestamps (and hence the checksum themselves) are always the same
        self._old_get_timestamp = _ValidHDU._get_timestamp
        _ValidHDU._get_timestamp = lambda self: "2013-12-20T13:36:10"

    def teardown_method(self):
        super().teardown_method()
        warnings.filters = self._oldfilters
        _ValidHDU._get_timestamp = self._old_get_timestamp


class TestChecksumFunctions(BaseChecksumTests):
    # All checksums have been verified against CFITSIO

    def test_sample_file(self):
        hdul = fits.open(self.data("checksum.fits"), checksum=True)
        assert hdul._read_all
        hdul.close()

    def test_image_create(self):
        n = np.arange(100, dtype=np.int64)
        hdu = fits.PrimaryHDU(n)
        hdu.writeto(self.temp("tmp.fits"), overwrite=True, checksum=True)
        with fits.open(self.temp("tmp.fits"), checksum=True) as hdul:
            assert (hdu.data == hdul[0].data).all()
            assert "CHECKSUM" in hdul[0].header
            assert "DATASUM" in hdul[0].header

            if not sys.platform.startswith("win32"):
                # The checksum ends up being different on Windows, possibly due
                # to slight floating point differences
                assert hdul[0].header["CHECKSUM"] == "ZHMkeGKjZGKjbGKj"
                assert hdul[0].header["DATASUM"] == "4950"

    def test_scaled_data(self):
        with fits.open(self.data("scale.fits")) as hdul:
            orig_data = hdul[0].data.copy()
            hdul[0].scale("int16", "old")
            hdul.writeto(self.temp("tmp.fits"), overwrite=True, checksum=True)
            with fits.open(self.temp("tmp.fits"), checksum=True) as hdul1:
                assert (hdul1[0].data == orig_data).all()
                assert "CHECKSUM" in hdul1[0].header
                assert hdul1[0].header["CHECKSUM"] == "cUmaeUjZcUjacUjW"
                assert "DATASUM" in hdul1[0].header
                assert hdul1[0].header["DATASUM"] == "1891563534"

    def test_scaled_data_auto_rescale(self):
        """
        Regression test for
        https://github.com/astropy/astropy/issues/3883#issuecomment-115122647

        Ensure that when scaled data is automatically rescaled on
        opening/writing a file that the checksum and datasum are computed for
        the rescaled array.
        """

        with fits.open(self.data("scale.fits")) as hdul:
            # Write out a copy of the data with the rescaling applied
            hdul.writeto(self.temp("rescaled.fits"))

        # Reopen the new file and save it back again with a checksum
        with fits.open(self.temp("rescaled.fits")) as hdul:
            hdul.writeto(self.temp("rescaled2.fits"), overwrite=True, checksum=True)

        # Now do like in the first writeto but use checksum immediately
        with fits.open(self.data("scale.fits")) as hdul:
            hdul.writeto(self.temp("rescaled3.fits"), checksum=True)

        # Also don't rescale the data but add a checksum
        with fits.open(self.data("scale.fits"), do_not_scale_image_data=True) as hdul:
            hdul.writeto(self.temp("scaled.fits"), checksum=True)

        # Must used nested with statements to support older Python versions
        # (but contextlib.nested is not available in newer Pythons :(
        with fits.open(self.temp("rescaled2.fits")) as hdul1:
            with fits.open(self.temp("rescaled3.fits")) as hdul2:
                with fits.open(self.temp("scaled.fits")) as hdul3:
                    hdr1 = hdul1[0].header
                    hdr2 = hdul2[0].header
                    hdr3 = hdul3[0].header
                    assert hdr1["DATASUM"] == hdr2["DATASUM"]
                    assert hdr1["CHECKSUM"] == hdr2["CHECKSUM"]
                    assert hdr1["DATASUM"] != hdr3["DATASUM"]
                    assert hdr1["CHECKSUM"] != hdr3["CHECKSUM"]

    def test_uint16_data(self):
        checksums = [
            ("aDcXaCcXaCcXaCcX", "0"),
            ("oYiGqXi9oXiEoXi9", "1746888714"),
            ("VhqQWZoQVfoQVZoQ", "0"),
            ("4cPp5aOn4aOn4aOn", "0"),
            ("8aCN8X9N8aAN8W9N", "1756785133"),
            ("UhqdUZnbUfnbUZnb", "0"),
            ("4cQJ5aN94aNG4aN9", "0"),
        ]
        with fits.open(self.data("o4sp040b0_raw.fits"), uint=True) as hdul:
            hdul.writeto(self.temp("tmp.fits"), overwrite=True, checksum=True)
            with fits.open(self.temp("tmp.fits"), uint=True, checksum=True) as hdul1:
                for idx, (hdu_a, hdu_b) in enumerate(zip(hdul, hdul1)):
                    if hdu_a.data is None or hdu_b.data is None:
                        assert hdu_a.data is hdu_b.data
                    else:
                        assert (hdu_a.data == hdu_b.data).all()

                    assert "CHECKSUM" in hdul[idx].header
                    assert hdul[idx].header["CHECKSUM"] == checksums[idx][0]
                    assert "DATASUM" in hdul[idx].header
                    assert hdul[idx].header["DATASUM"] == checksums[idx][1]

    def test_groups_hdu_data(self):
        imdata = np.arange(100.0)
        imdata.shape = (10, 1, 1, 2, 5)
        pdata1 = np.arange(10) + 0.1
        pdata2 = 42
        x = fits.hdu.groups.GroupData(
            imdata, parnames=["abc", "xyz"], pardata=[pdata1, pdata2], bitpix=-32
        )
        hdu = fits.GroupsHDU(x)
        hdu.writeto(self.temp("tmp.fits"), overwrite=True, checksum=True)
        with fits.open(self.temp("tmp.fits"), checksum=True) as hdul:
            assert comparerecords(hdul[0].data, hdu.data)
            assert "CHECKSUM" in hdul[0].header
            assert hdul[0].header["CHECKSUM"] == "3eDQAZDO4dDOAZDO"
            assert "DATASUM" in hdul[0].header
            assert hdul[0].header["DATASUM"] == "2797758084"

    def test_binary_table_data(self):
        a1 = np.array(["NGC1001", "NGC1002", "NGC1003"])
        a2 = np.array([11.1, 12.3, 15.2])
        col1 = fits.Column(name="target", format="20A", array=a1)
        col2 = fits.Column(name="V_mag", format="E", array=a2)
        cols = fits.ColDefs([col1, col2])
        tbhdu = fits.BinTableHDU.from_columns(cols)
        tbhdu.writeto(self.temp("tmp.fits"), overwrite=True, checksum=True)
        with fits.open(self.temp("tmp.fits"), checksum=True) as hdul:
            assert comparerecords(tbhdu.data, hdul[1].data)
            assert "CHECKSUM" in hdul[0].header
            assert hdul[0].header["CHECKSUM"] == "D8iBD6ZAD6fAD6ZA"
            assert "DATASUM" in hdul[0].header
            assert hdul[0].header["DATASUM"] == "0"
            assert "CHECKSUM" in hdul[1].header
            assert hdul[1].header["CHECKSUM"] == "aD1Oa90MaC0Ma90M"
            assert "DATASUM" in hdul[1].header
            assert hdul[1].header["DATASUM"] == "1062205743"

    def test_variable_length_table_data(self):
        c1 = fits.Column(
            name="var",
            format="PJ()",
            array=np.array([[45.0, 56], np.array([11, 12, 13])], "O"),
        )
        c2 = fits.Column(name="xyz", format="2I", array=[[11, 3], [12, 4]])
        tbhdu = fits.BinTableHDU.from_columns([c1, c2])
        tbhdu.writeto(self.temp("tmp.fits"), overwrite=True, checksum=True)
        with fits.open(self.temp("tmp.fits"), checksum=True) as hdul:
            assert comparerecords(tbhdu.data, hdul[1].data)
            assert "CHECKSUM" in hdul[0].header
            assert hdul[0].header["CHECKSUM"] == "D8iBD6ZAD6fAD6ZA"
            assert "DATASUM" in hdul[0].header
            assert hdul[0].header["DATASUM"] == "0"
            assert "CHECKSUM" in hdul[1].header
            assert hdul[1].header["CHECKSUM"] == "YIGoaIEmZIEmaIEm"
            assert "DATASUM" in hdul[1].header
            assert hdul[1].header["DATASUM"] == "1507485"

    def test_ascii_table_data(self):
        a1 = np.array(["abc", "def"])
        r1 = np.array([11.0, 12.0])
        c1 = fits.Column(name="abc", format="A3", array=a1)
        # This column used to be E format, but the single-precision float lost
        # too much precision when scaling so it was changed to a D
        c2 = fits.Column(name="def", format="D", array=r1, bscale=2.3, bzero=0.6)
        c3 = fits.Column(name="t1", format="I", array=[91, 92, 93])
        x = fits.ColDefs([c1, c2, c3])
        hdu = fits.TableHDU.from_columns(x)
        hdu.writeto(self.temp("tmp.fits"), overwrite=True, checksum=True)
        with fits.open(self.temp("tmp.fits"), checksum=True) as hdul:
            assert comparerecords(hdu.data, hdul[1].data)
            assert "CHECKSUM" in hdul[0].header
            assert hdul[0].header["CHECKSUM"] == "D8iBD6ZAD6fAD6ZA"
            assert "DATASUM" in hdul[0].header
            assert hdul[0].header["DATASUM"] == "0"

            if not sys.platform.startswith("win32"):
                # The checksum ends up being different on Windows, possibly due
                # to slight floating point differences
                assert "CHECKSUM" in hdul[1].header
                assert hdul[1].header["CHECKSUM"] == "3rKFAoI94oICAoI9"
                assert "DATASUM" in hdul[1].header
                assert hdul[1].header["DATASUM"] == "1914653725"

    def test_open_with_no_keywords(self):
        hdul = fits.open(self.data("arange.fits"), checksum=True)
        hdul.close()

    def test_append(self):
        hdul = fits.open(self.data("tb.fits"))
        hdul.writeto(self.temp("tmp.fits"), overwrite=True)
        n = np.arange(100)
        fits.append(self.temp("tmp.fits"), n, checksum=True)
        hdul.close()
        hdul = fits.open(self.temp("tmp.fits"), checksum=True)
        assert hdul[0]._checksum is None
        hdul.close()

    def test_writeto_convenience(self):
        n = np.arange(100)
        fits.writeto(self.temp("tmp.fits"), n, overwrite=True, checksum=True)
        hdul = fits.open(self.temp("tmp.fits"), checksum=True)
        self._check_checksums(hdul[0])
        hdul.close()

    def test_hdu_writeto(self):
        n = np.arange(100, dtype="int16")
        hdu = fits.ImageHDU(n)
        hdu.writeto(self.temp("tmp.fits"), checksum=True)
        hdul = fits.open(self.temp("tmp.fits"), checksum=True)
        self._check_checksums(hdul[0])
        hdul.close()

    def test_hdu_writeto_existing(self):
        """
        Tests that when using writeto with checksum=True, a checksum and
        datasum are added to HDUs that did not previously have one.

        Regression test for https://github.com/spacetelescope/PyFITS/issues/8
        """

        with fits.open(self.data("tb.fits")) as hdul:
            hdul.writeto(self.temp("test.fits"), checksum=True)

        with fits.open(self.temp("test.fits")) as hdul:
            assert "CHECKSUM" in hdul[0].header
            # These checksums were verified against CFITSIO
            assert hdul[0].header["CHECKSUM"] == "7UgqATfo7TfoATfo"
            assert "DATASUM" in hdul[0].header
            assert hdul[0].header["DATASUM"] == "0"
            assert "CHECKSUM" in hdul[1].header
            assert hdul[1].header["CHECKSUM"] == "99daD8bX98baA8bU"
            assert "DATASUM" in hdul[1].header
            assert hdul[1].header["DATASUM"] == "1829680925"

    def test_datasum_only(self):
        n = np.arange(100, dtype="int16")
        hdu = fits.ImageHDU(n)
        hdu.writeto(self.temp("tmp.fits"), overwrite=True, checksum="datasum")
        with fits.open(self.temp("tmp.fits"), checksum=True) as hdul:
            if not (hasattr(hdul[0], "_datasum") and hdul[0]._datasum):
                pytest.fail(msg="Missing DATASUM keyword")

            if not (hasattr(hdul[0], "_checksum") and not hdul[0]._checksum):
                pytest.fail(msg="Non-empty CHECKSUM keyword")

    def test_open_update_mode_preserve_checksum(self):
        """
        Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/148 where
        checksums are being removed from headers when a file is opened in
        update mode, even though no changes were made to the file.
        """

        self.copy_file("checksum.fits")

        with fits.open(self.temp("checksum.fits")) as hdul:
            data = hdul[1].data.copy()

        hdul = fits.open(self.temp("checksum.fits"), mode="update")
        hdul.close()

        with fits.open(self.temp("checksum.fits")) as hdul:
            assert "CHECKSUM" in hdul[1].header
            assert "DATASUM" in hdul[1].header
            assert comparerecords(data, hdul[1].data)

    def test_open_update_mode_update_checksum(self):
        """
        Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/148, part
        2.  This ensures that if a file contains a checksum, the checksum is
        updated when changes are saved to the file, even if the file was opened
        with the default of checksum=False.

        An existing checksum and/or datasum are only stripped if the file is
        opened with checksum='remove'.
        """

        self.copy_file("checksum.fits")
        with fits.open(self.temp("checksum.fits")) as hdul:
            header = hdul[1].header.copy()
            data = hdul[1].data.copy()

        with fits.open(self.temp("checksum.fits"), mode="update") as hdul:
            hdul[1].header["FOO"] = "BAR"
            hdul[1].data[0]["TIME"] = 42

        with fits.open(self.temp("checksum.fits")) as hdul:
            header2 = hdul[1].header
            data2 = hdul[1].data
            assert header2[:-3] == header[:-2]
            assert "CHECKSUM" in header2
            assert "DATASUM" in header2
            assert header2["FOO"] == "BAR"
            assert (data2["TIME"][1:] == data["TIME"][1:]).all()
            assert data2["TIME"][0] == 42

        with fits.open(
            self.temp("checksum.fits"), mode="update", checksum="remove"
        ) as hdul:
            pass

        with fits.open(self.temp("checksum.fits")) as hdul:
            header2 = hdul[1].header
            data2 = hdul[1].data
            assert header2[:-1] == header[:-2]
            assert "CHECKSUM" not in header2
            assert "DATASUM" not in header2
            assert header2["FOO"] == "BAR"
            assert (data2["TIME"][1:] == data["TIME"][1:]).all()
            assert data2["TIME"][0] == 42

    def test_overwrite_invalid(self):
        """
        Tests that invalid checksum or datasum are overwritten when the file is
        saved.
        """

        reffile = self.temp("ref.fits")
        with fits.open(self.data("tb.fits")) as hdul:
            hdul.writeto(reffile, checksum=True)

        testfile = self.temp("test.fits")
        with fits.open(self.data("tb.fits")) as hdul:
            hdul[0].header["DATASUM"] = "1       "
            hdul[0].header["CHECKSUM"] = "8UgqATfo7TfoATfo"
            hdul[1].header["DATASUM"] = "2349680925"
            hdul[1].header["CHECKSUM"] = "11daD8bX98baA8bU"
            hdul.writeto(testfile)

        with fits.open(testfile) as hdul:
            hdul.writeto(self.temp("test2.fits"), checksum=True)

        with fits.open(self.temp("test2.fits")) as hdul:
            with fits.open(reffile) as ref:
                assert "CHECKSUM" in hdul[0].header
                # These checksums were verified against CFITSIO
                assert hdul[0].header["CHECKSUM"] == ref[0].header["CHECKSUM"]
                assert "DATASUM" in hdul[0].header
                assert hdul[0].header["DATASUM"] == "0"
                assert "CHECKSUM" in hdul[1].header
                assert hdul[1].header["CHECKSUM"] == ref[1].header["CHECKSUM"]
                assert "DATASUM" in hdul[1].header
                assert hdul[1].header["DATASUM"] == ref[1].header["DATASUM"]

    def _check_checksums(self, hdu):
        if not (hasattr(hdu, "_datasum") and hdu._datasum):
            pytest.fail(msg="Missing DATASUM keyword")

        if not (hasattr(hdu, "_checksum") and hdu._checksum):
            pytest.fail(msg="Missing CHECKSUM keyword")



"""
A simple class to manage a piece of global science state.  See
:ref:`astropy:config-developer` for more details.
"""

__all__ = ["ScienceState"]


class _ScienceStateContext:
    def __init__(self, parent, value):
        self._value = value
        self._parent = parent

    def __enter__(self):
        pass

    def __exit__(self, type, value, tb):
        self._parent._value = self._value

    def __repr__(self):
        # Ensure we have a single-line repr, just in case our
        # value is not something simple like a string.
        value_repr, lb, _ = repr(self._parent._value).partition("\n")
        if lb:
            value_repr += "..."
        return f"<ScienceState {self._parent.__name__}: {value_repr}>"


class ScienceState:
    """
    Science state subclasses are used to manage global items that can
    affect science results.  Subclasses will generally override
    `validate` to convert from any of the acceptable inputs (such as
    strings) to the appropriate internal objects, and set an initial
    value to the ``_value`` member so it has a default.

    Examples
    --------
    ::

        class MyState(ScienceState):
            @classmethod
            def validate(cls, value):
                if value not in ('A', 'B', 'C'):
                    raise ValueError("Must be one of A, B, C")
                return value
    """

    def __init__(self):
        raise RuntimeError("This class is a singleton.  Do not instantiate.")

    @classmethod
    def get(cls):
        """
        Get the current science state value.
        """
        return cls.validate(cls._value)

    @classmethod
    def set(cls, value):
        """Set the current science state value."""
        # Create context with current value
        ctx = _ScienceStateContext(cls, cls._value)

        # Set new value
        value = cls.validate(value)
        cls._value = value

        # Return context manager
        return ctx

    @classmethod
    def validate(cls, value):
        """
        Validate the value and convert it to its native type, if
        necessary.
        """
        return value



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This package reads and writes data formats used by the Virtual
Observatory (VO) initiative, particularly the VOTable XML format.
"""

from astropy import config as _config

from .exceptions import (
    IOWarning,
    UnimplementedWarning,
    VOTableChangeWarning,
    VOTableSpecError,
    VOTableSpecWarning,
    VOWarning,
)
from .table import from_table, is_votable, parse, parse_single_table, validate, writeto

__all__ = [
    "Conf",
    "conf",
    "parse",
    "parse_single_table",
    "validate",
    "from_table",
    "is_votable",
    "writeto",
    "VOWarning",
    "VOTableChangeWarning",
    "VOTableSpecWarning",
    "UnimplementedWarning",
    "IOWarning",
    "VOTableSpecError",
]

VERIFY_OPTIONS = ["ignore", "warn", "exception"]  # First one is default


class Conf(_config.ConfigNamespace):
    """
    Configuration parameters for `astropy.io.votable`.
    """

    verify = _config.ConfigItem(
        VERIFY_OPTIONS,
        "Can be 'exception' (treat fixable violations of the VOTable spec as "
        "exceptions), 'warn' (show warnings for VOTable spec violations), or "
        "'ignore' (silently ignore VOTable spec violations)",
    )


conf = Conf()
